'' =================================================================================================
''
''   File....... jm_fdsx.spin
''   Purpose.... True-mode, minimalist serial driver
''               -- external buffers to control RAM use
''   Author..... Jon "JonnyMac" McPhalen
''               Copyright (c) 2022 Jon McPhalen
''               -- based on code by Chip Gracey
''               -- see below for terms of use
''   E-mail..... jon.mcphalen@gmail.com
''   Started....
''   Updated.... 29 OCT 2022
''
'' =================================================================================================


con { fixed io pins }

  RX_PGM   = 31  { I }                                          ' serial / programming
  TX_PGM   = 30  { O }

  EE_SDA   = 29  { I/O }                                        ' i2c / eeprom
  EE_SCL   = 28  { I/O }


var

  long  cog                                                     ' cog (+1) running uart code

  long  ms1                                                     ' ticks in 1ms


  ' 11 longs, MUST be contiguous
  ' -- do not change order

  long  rxhead
  long  rxtail
  long  txhead
  long  txtail

  long  rxpin                                                   ' from caller
  long  rxbufsize
  long  rxaddr
  long  txpin
  long  txbufsize
  long  txaddr
  long  bitticks


pub null

'' This is not an application


pub start(rxp, rxsize, p_rxbuf, txp, txsize, p_txbuf, baud)

'' Start serial driver (uses a cog, returns 1 to 8 if successful)
'' -- true mode only
'' -- rxp      = recieve pin (0..31)
'' -- rxsize   = size of rx buffer in bytes
'' -- p_rxbuf  = pointer to rx buffer
'' -- txp      = transmit pin (0..31)
'' -- txsize   = size of tx buffer in bytes
'' -- p_txbuf  = pointer to tx buffer
'' -- baud     = serial baud (e.g. 9600)

  stop                                                          ' stop if running

  longmove(@rxpin, @rxp, 6)                                     ' copy connection parameters
  bitticks := clkfreq / baud                                    ' ticks in one bit period

  ms1 := clkfreq / 1000                                         ' ticks in 1ms

  cog := cognew(@fds_uart, @rxhead) + 1                         ' start the fds uart cog

  return cog


pub stop

'' Stop serial driver -- frees a cog

  if (cog)
    cogstop(cog-1)
    cog := 0
    longfill(@rxhead, 0, 11)


pub available : result

'' Returns # of bytes waiting in rx buffer

  result := rxhead - rxtail
  if (result < 0)
    result += rxbufsize


pub rxflush

'' Flush receive buffer

  repeat while (rxcheck => 0)


pub rxcheck : rxbyte

'' Check if byte received (never waits)
'' -- returns -1 if no byte received, $00..$FF if byte available

  rxbyte := -1                                                  ' assume no byte ready

  if (rxtail <> rxhead)                                         ' if byte(s) in buffer
    rxbyte := byte[rxaddr][rxtail]                              ' get next available
    if (++rxtail == rxbufsize)                                  ' increment pointer, wrap if needed
      rxtail := 0


pub rxtime(ms) : rxbyte | t

'' Wait ms milliseconds for a byte to be received
'' -- returns -1 if no byte received, $00..$FF if byte available

  t := cnt
  repeat until ((rxbyte := rxcheck) => 0) or (((cnt-t) / ms1) => ms)

  return rxbyte


pub rx : rxbyte

'' Receive byte (may wait for byte)
'' -- returns $00..$FF

  repeat while ((rxbyte := rxcheck) < 0)

  return rxbyte


pub txflush

'' Wait for transmit buffer to empty, then wait for last byte to transmit
'' -- call before stopping cog

  repeat until (txtail == txhead)
  repeat 11
    waitcnt(cnt + bitticks)


pub tx(txbyte) : space

'' Send byte
'' -- may wait for room in buffer

  repeat while (space =< 1)                                     ' verify/wait for room in buffer
    if (txhead => txtail)
      space := txbufsize - (txhead - txtail)
    else
      space := txtail - txhead

  byte[txaddr][txhead] := txbyte                                ' add new byte to buffer
  if (++txhead == txbufsize)                                    ' update head pointer
    txhead := 0


pub str(p_zstr)

'' Send zero-terminated string at p_zstr

  repeat strsize(p_zstr)
    tx(byte[p_zstr++])


dat { pasm driver -- true mode only }

' Derived from code by Chip Gracey, Parallax
' -- removed open and inverted modes
' -- modified setup and buffer size handling

                        org     0

fds_uart                mov     t1, par                         ' get structure address
                        add     t1, #16                         ' skip past heads and tails (4 longs)

get_parms               movd    :read, #rxmask                  ' location of cog parameters -> :read(dest)
                        mov     t2, #7                          ' transfer 7 parameters
:read                   rdlong  0-0, t1                         ' copy parameter from hub to cog
                        add     t1, #4                          ' next hub element
                        add     :read, INC_DEST                 ' next cog element
                        djnz    t2, #:read                      ' done?

                        mov     t1, #1                          ' convert rx pin to mask
                        shl     t1, rxmask
                        mov     rxmask, t1

                        mov     t1, #1                          ' convert tx pin to mask
                        shl     t1, txmask
                        mov     txmask, t1
                        or      outa, txmask                    ' set to idle
                        or      dira, txmask

                        mov     txcode, #transmit               ' initialize ping-pong multitasking


' -----------------
'  Receive Process
' -----------------
'
receive                 jmpret  rxcode, txcode                  ' run a chunk of transmit code, then return

                        test    rxmask, ina             wc      ' sample rx pin
        if_c            jmp     #receive

                        mov     rxbits, #9                      ' ready to receive byte
                        mov     rxcnt, bittix
                        shr     rxcnt, #1
                        add     rxcnt, cnt

:bit                    add     rxcnt, bittix                   ' ready next bit period

:wait                   jmpret  rxcode, txcode                  ' run a chuck of transmit code, then return

                        mov     t1, rxcnt                       ' check if bit receive period done
                        sub     t1, cnt
                        cmps    t1, #0                  wc
        if_nc           jmp     #:wait

                        test    rxmask, ina             wc      ' receive bit on rx pin
                        rcr     rxdata, #1
                        djnz    rxbits, #:bit

                        shr     rxdata, #32-9                   ' justify lsb for wrbyte

                        rdlong  t2, par                         ' read head index
                        add     t2, rxhub                       ' convert to head pointer
                        wrbyte  rxdata, t2                      ' write the new byte

                        sub     t2, rxhub                       ' back to head index
                        add     t2, #1                          ' increment
                        cmpsub  t2, rxlen                       ' if past end, back to 0
                        wrlong  t2, par

                        jmp     #receive                        ' byte done, receive next byte


' ------------------
'  Transmit Process
' ------------------
'
transmit                jmpret  txcode, rxcode                  ' run a chunk of receive code, then return

                        mov     t1, par                         ' check for head <> tail
                        add     t1, #8
                        rdlong  t2, t1                          ' t1 = @txhead, t2 := txhead
                        add     t1, #4
                        rdlong  t3, t1                          ' t1 = @txtail, t3 := txtail
                        cmp     t2, t3                  wz      ' equal?
        if_z            jmp     #transmit                       ' if yes, check again

                        add     t3, txhub                       ' get byte and inc tail
                        rdbyte  txdata, t3
                        sub     t3, txhub
                        add     t3, #1
                        cmpsub  t3, txlen
                        wrlong  t3, t1

                        or      txdata, STOP_BITS               ' add stop bit(s)
                        shl     txdata, #1                      ' add start bit
                        mov     txbits, #11                     ' bits = start + 8 + 2 stop
                        mov     txcnt, cnt

:bit                    shr     txdata, #1              wc      ' get bit
                        muxc    outa, txmask                    ' output to tx line
                        add     txcnt, bittix                   ' ready next cnt

:wait                   jmpret  txcode, rxcode                  ' run a chunk of receive code, then return

                        mov     t1, txcnt                       ' check if bit transmit period done
                        sub     t1, cnt
                        cmps    t1, #0                  wc
        if_nc           jmp     #:wait

                        djnz    txbits, #:bit                   ' another bit to transmit?

                        jmp     #transmit                       ' byte done, transmit next byte

' --------------------------------------------------------------------------------------------------

INC_DEST                long    1 << 9                          ' to increment D field

STOP_BITS               long    $FFFF_FF00

rxmask                  res     1                               ' mask for rx pin
rxlen                   res     1                               ' length of rx buffer
rxhub                   res     1                               ' hub address of rxbuffer                             '
txmask                  res     1                               ' mask for tx pin
txlen                   res     1                               ' length of tx buffer
txhub                   res     1                               ' hub address of txbuffer
bittix                  res     1

rxdata                  res     1                               ' byte received
rxbits                  res     1                               ' bit counter for rx
rxcnt                   res     1                               ' bit timer for rx
rxcode                  res     1                               ' cog pointer for rx process

txdata                  res     1                               ' byte to transmit
txbits                  res     1                               ' bit counter for tx
txcnt                   res     1                               ' bit timer for tx
txcode                  res     1                               ' cog pointer for tx process

t1                      res     1                               ' working registers
t2                      res     1
t3                      res     1

                        fit     496


con { license }

{{

  Terms of Use: MIT License

  Permission is hereby granted, free of charge, to any person obtaining a copy of this
  software and associated documentation files (the "Software"), to deal in the Software
  without restriction, including without limitation the rights to use, copy, modify,
  merge, publish, distribute, sublicense, and/or sell copies of the Software, and to
  permit persons to whom the Software is furnished to do so, subject to the following
  conditions:

  The above copyright notice and this permission notice shall be included in all copies
  or substantial portions of the Software.

  THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED,
  INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A
  PARTICULAR PURPOSE AND NON-INFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT
  HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF
  CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE
  OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.

}}