'' =================================================================================================
''
''   File....... jm_nextion.spin
''   Purpose....
''   Author..... Jon McPhalen
''               Copyright (c) 2022 Jon McPhalen
''               -- see below for terms of use
''   E-mail..... jon.mcphalen@gmail.com
''   Started....
''   Updated.... 20 DEC 2022
''
''   {$P1}
''
'' =================================================================================================

{{

   Note: JM-style custom command/value strings from Nextion should be in the form:

     command=value\r

   -- no spaces
   -- if value is negative, "-" must be first character after "="
   -- value can be binary with "%" or hex with "$"
   -- to append [CR] in Nextion use

   Example: To notify the Propeller that the Nextion has switched to page 3, put this code
            into the post-initialize event for the page.

     prints "page=3",6
     prints 0x0D,1

   (Do not use \r in Nextion string as it embeds $0D and $0A)

   The command parser can deal with standard nextion messages but will return a negative
   length.

}}


con { fixed io pins }

  RX_PGM   = 31  { I }                                          ' serial / programming
  TX_PGM   = 30  { O }

  EE_SDA   = 29  { I/O }                                        ' i2c / eeprom
  EE_SCL   = 28  { I/O }


con { colors }

' Color constants understood by Nextion

  BLACK  =     0
  BLUE   =    31
  BROWN  = 48192
  GREEN  =  2016
  YELLOW = 65504
  RED    = 63488
  GRAY   = 33840
  WHITE  = 65535


con

  #true,  ON, OFF
  #false, NO, YES

  #0, NXT_MSG, CUST_MSG                                         '

  HMI_RXBUFSZ = 96
  HMI_TXBUFSZ = 48
  HMI_MSGSZ   = 32

  SEP = "="                                                     ' command/value separator
  CR  =  13                                                     ' carriage return
  EOF =   0                                                     ' end of string


obj

  fdsx : "jm_fdsx"                                              ' * mimimalist full-duplex serial

' * uses cog when loaded


dat

  s_connect     byte    "connect", 0
  s_get         byte    "get ", 0
  s_txt         byte    ".txt", 0
  s_comok       byte    "comok ", 0


var { serial buffers }

  byte  rxbuf[HMI_RXBUFSZ]                                      ' hmi serial buffers
  byte  txbuf[HMI_TXBUFSZ]


var

  long  cog
  long  ms1ticks

  byte  buffer[HMI_MSGSZ]                                       ' string buffer


pub null

'' This is not an application


pub start(rxp, txp, baud)  : result

'' Start serial driver for Nextion
'' -- rxp      = recieve pin (0..31)
'' -- txp      = transmit pin (0..31)
'' -- baud     = serial baud (e.g. 9600)

  stop

  cog := fdsx.start(rxp, HMI_RXBUFSZ, @rxbuf, txp, HMI_TXBUFSZ, @txbuf, baud)

  ms1ticks := clkfreq / 1000

  return cog


pub stop

'' Stops Nextion serial driver if running

  if (cog)
    fdsx.stop


pub reset(delay, page) : result

'' Reset Nextion, hold, switch to desired page
'' -- minimum hold is 500 ms to allow post-reset response
''    * auto response cleared
'' -- use page >0 to switch to different page

  fdsx.str(string("rest"))
  terminate
  delay := 1000 #> delay
  waitms(delay)
  fdsx.rxflush

  if (page > 0)
    return set_page(page)


pub connect(p_str) | b

'' Requests and returns raw connect string from Nextion
'' -- fields are separated by commas
'' -- p_str must be big enough to hold return string (70 bytes)
'' -- reference : https://nextion.ca/documentation/nextion-upload-protocol-v1-1/

  command(@s_Connect)
  waitms(150)                                                   ' allow for 9600 baud

  repeat
    b := fdsx.rxcheck
    if (b => 0)
      byte[p_str++] := b
    else
      byte[p_str] := 0
      quit


pub has_connection(p_str) : result | check[2]

'' Looks for "comok" at start of p_str
'' -- should be there if p_str is connection response

  bytemove(@check, p_str, 6)
  check.byte[6] := 0

  return strcomp(@check, @s_comok)


pub get_touch(p_src) : result

'' Return true of Nextion has touch fe
'' -- use connect() to get conneciton details first

  if (has_connection(p_src))
    return (byte[p_src][6] == "1")
  else
    return false


pub get_model(p_src, p_dest)

'' Parse Nextion model from connection response string (p_src)
'' -- use connect() to get conneciton details first

  if (has_connection(p_src))
    extract_field(p_src, 2, p_dest)
  else
    byte[p_dest] := 0


pub get_firmware(p_src) : result

'' Parse Nextion firmware version from connection response string (p_src)
'' -- use connect() to get conneciton details first

  return extract_fvalue(p_src, 3)


pub get_serial_number(p_src, p_dest)

'' Parse Nextion serial number from connection response string (p_src)
'' -- use connect() to get conneciton details first

  if (has_connection(p_src))
    extract_field(p_src, 5, p_dest)
  else
    byte[p_dest] := 0


pub get_flash_size(p_src) : result

'' Parse Nextion flash size from connection response string (p_src)
'' -- use connect() to get conneciton details first

  return extract_fvalue(p_src, 6)


pri extract_field(p_src, n, p_dest) : found | b

'' Extract string from comma-delimited field

  repeat
    b := byte[p_src++]
    if (b == 0) or (b == 255)
      byte[p_dest] := 0
      return
    else
      if (b == ",")
        if (++found == n)
          quit

  repeat
    b := byte[p_src++]
    if (b == 0) or (b == ",") or (b == 255)
      byte[p_dest] := 0
      return
    else
      byte[p_dest++] := b


pri extract_fvalue(p_src, n) : result | b

'' Extract decimal value from comma-delimited field

  repeat
    b := byte[p_src++]
    if (b == 0) or (b == 255)
      return 0
    else
      if (b == ",")
        if (++result == n)
          quit

  result := 0

  repeat
    b := byte[p_src++]
    case b
      "0".."9" : result := result * 10 + (b - "0")
      other    : return


pub brightness(level)

'' Sets display backlight brightness

  set_value(string("dim"), 0 #> level <# 100)


pub command(p_str)

'' Sends p_str as completed command

  fdsx.rxflush
  fdsx.str(p_str)
  terminate                                                     ' \xFF\xFF\xFF
  fdsx.txflush


pub set_page(dp) : result

'' Set Nextion page
'' -- returns true if successful

  fdsx.rxflush
  fdsx.str(string("page "))
  dec2asc(dp)
  terminate
  fdsx.txflush
  waitms(10)

  if (fdsx.available == 0)
    return true
  else
    fdsx.rxflush
    return false


pub set_value(s_obj, value) : result

'' Send value to object defined by s_obj
'' -- screen objects must include target field, e.g, "gauge.pco" (gauge pointer color)

  fdsx.rxflush
  fdsx.str(s_obj)
  fdsx.tx("=")
  dec2asc(value)
  terminate
  fdsx.txflush
  waitms(10)

  if (fdsx.available == 0)
    return true
  else
    fdsx.rxflush
    return false


pub set_text(s_obj, p_str) : result

'' Send string to text field of object defined by s_obj
'' -- do not include '.txt' in object name

  fdsx.rxflush
  fdsx.str(s_obj)
  fdsx.str(@s_txt)
  fdsx.tx("=")
  fdsx.tx($22)
  fdsx.str(p_str)
  fdsx.tx($22)
  terminate
  fdsx.txflush
  waitms(10)

  if (fdsx.available == 0)
    return true
  else
    fdsx.rxflush
    return false


pub set_component_value(id, p_field, value)

'' Set field for component id to value
'' -- applies to current page

  fdsx.rxflush
  fdsx.str(string("b["))
  dec2asc(id)
  fdsx.str(string("]."))
  fdsx.str(p_field)
  fdsx.tx("=")
  dec2asc(value)
  terminate
  fdsx.txflush
  waitms(10)

  if (fdsx.available == 0)
    return true
  else
    fdsx.rxflush
    return false


pub set_component_str(id, p_field, p_str)

'' Set field for component id to p_str (for .txt fields)
'' -- applies to current page

  fdsx.rxflush
  fdsx.str(string("b["))
  dec2asc(id)
  fdsx.str(string("]."))
  fdsx.str(p_field)
  fdsx.tx("=")
  fdsx.tx($22)
  fdsx.str(p_str)
  fdsx.tx($22)
  terminate
  fdsx.txflush
  waitms(10)

  if (fdsx.available == 0)
    return true
  else
    fdsx.rxflush
    return false


pub click(s_obj, value) : result

'' Run click event code on defined by s_obj

  fdsx.rxflush
  fdsx.str(string("click "))
  fdsx.str(s_obj)
  fdsx.tx(",")
  if (value)
    fdsx.tx("1")
  else
    fdsx.tx("0")
  terminate
  fdsx.txflush
  waitms(10)

  if (fdsx.available == 0)
    return true
  else
    fdsx.rxflush
    return false


pub get_value(s_obj) : result | x, b

'' Get numeric value from Nextion object
'' -- s_obj is Nextion string name of object (e.g.,"dp" for page)
'' -- negx returned on error

  fdsx.rxflush                                                  ' clear old rx
  fdsx.str(@s_get)                                              ' get item
  fdsx.str(s_obj)
  terminate
  fdsx.txflush                                                  ' let tx finish
  waitms(10)                                                    ' let Nextion respond

  repeat x from 0 to 7                                          ' wait for response
    b := fdsx.rxtime(5)
    if (b => 0)
      buffer.byte[x] := b
    else
      return negx                                               ' no/fouled response

  if (buffer.byte[0] == $71)                                    ' proper header?
    bytemove(@result, @buffer.byte[1], 4)                       ' move value to result
  else
    fdsx.rxflush
    result := negx


pub get_text(s_obj, p_dest) : len | b

'' Get string from text field of object defined by s_obj; return length
'' -- output to byte array at p_dest
'' -- do not include '.txt' in object name

  fdsx.rxflush                                                  ' clear old rx
  fdsx.str(@s_get)                                              ' get item
  fdsx.str(s_obj)
  fdsx.str(string(".txt"))
  terminate
  fdsx.txflush                                                  ' let tx finish
  waitms(10)                                                    ' let Nextion respond

  b := fdsx.rxtime(5)
  if (b == $70)                                                 ' good response
    repeat
      b := fdsx.rxtime(5)
      if (b > 0) and (b <> $FF)
        byte[p_dest][len++] := b                                ' build result string
      else
        byte[p_dest][len] := 0                                  ' terminate output string
        quit

  fdsx.rxflush


pub check_msg(p_dest, maxlen) : len | c, ffcnt

'' Receive message from Nextion into p_dest
'' -- usually in the form "command=value" + $0D
'' -- maxlen specifies maximum length to recieve
'' -- negative length indicates standard Nextion message

  bytefill(p_dest, 0, maxlen)                                   ' clear buffer

  ffcnt := 0                                                    ' $FF count (error message terminator)

  repeat maxlen-1
    c := fdsx.rxtime(5)                                         ' look for message
    case c
      CR, 32..126 :
        byte[p_dest][len++] := c                                ' add to string
        if (c == CR)                                            ' if CR we're done (custom)
          return

      $FF :
        byte[p_dest][len++] := c
        if (++ffcnt == 3)
          return -len                                           ' Nextion message

      EOF, -1 :
        return


pub extract_cmd(p_str) : p_cmd | x, b

'' Returns substring on the left side of "="

  clear_cmd

  repeat x from 0 to strsize(p_str)-1
    b := byte[p_str][x]
    if (b == SEP) or (b == CR) or (b == EOF)                    ' found end
      quit
    else
      buffer[x] := b

  return @buffer


pub last_cmd : p_str

'' Returns last command string or text request

  return @buffer


pub clear_cmd

'' Clears last command to prevent re-use

  bytefill(@buffer, 0, HMI_MSGSZ)


pub extract_value(p_str) : value | b

'' Returns value on the rights side of "="
'' -- for negative decimal values, 1st char after "=" must be "-"
'' -- hex values must lead with "$"
'' -- bin values must lead with "_"
'' -- the "_" is allowed between digits

  repeat
    b := byte[p_str++]                                          ' get character
    if (b == CR) or (b == EOF)                                  ' end?
      return
    elseif (b == SEP)                                           ' if separator found
      return get_val(p_str)                                     ' return value from string


pub get_val(p_str) : value | b

'' Determine numeric type and call ascii-to-value conversion
'' -- no spaces allowed after "=" in command/value string

  b := byte[p_str]
  case b
    "-", "0".."9" : return dec2val(p_str)
    "$"           : return hex2val(++p_str)                     ' skip $
    "%"           : return bin2val(++p_str)                     ' skip %


pub dec2val(p_str) : value | x, sign, b

'' Convert decimal string to long

  x := 0
  sign := 1

  repeat
    b := byte[p_str][x++]
    if (b == "-")
      if (x == 1)
        sign := -1
      else
        return value * sign

    elseif (b => "0") and (b =< "9")
      value := (value * 10) + (b - "0")

    elseif (b <> "_")
      return value * sign


pub hex2val(p_str) : value | b

'' Convert hex string to long

  if (byte[p_str] == "$")                                       ' remove indicator
    p_str++

  repeat
    b := byte[p_str++]
    if (b => "a") and (b =< "z")                                ' make uppercase
      b -= 32
    if (b => "0") and (b =< "9")
      value := (value << 4) + (b - "0")

    elseif (b => "A") and (b =< "F")
      value := (value << 4) + (b - "A") + 10

    elseif (b <> "_")
      return


pub bin2val(p_str) : value | b

'' Convert binary string to long

  if (byte[p_str] == "%")
    p_str++

  repeat
    b := byte[p_str++]
    if (b => "0") and (b =< "1")
      value := (value << 1) + (b - "0")
    elseif (b <> "_")
      return


con { helpers }

pub dec2asc(value) : zf | d, c

'' Send signed decimal (no formatting)
'' -- does not work with negx

  if (value < 0)
    fdsx.tx("-")
    -value

  d := 1_000_000_000

  repeat 10
    if (value => d)
      fdsx.tx("0" + value / d)
      value //= d
      zf := 1
    elseif (zf) or (d == 1)
      fdsx.tx("0")
    d /= 10


pub terminate

'' Nextion messages end with $FF $FF $FF

  repeat 3
    fdsx.tx($FF)


pri waitms(delay) : t

'' Delay in milliseconds

  t := cnt
  repeat while (delay > 0)
    waitcnt(t += ms1ticks)
    --delay


con { license }

{{

  Terms of Use: MIT License

  Permission is hereby granted, free of charge, to any person obtaining a copy of this
  software and associated documentation files (the "Software"), to deal in the Software
  without restriction, including without limitation the rights to use, copy, modify,
  merge, publish, distribute, sublicense, and/or sell copies of the Software, and to
  permit persons to whom the Software is furnished to do so, subject to the following
  conditions:

  The above copyright notice and this permission notice shall be included in all copies
  or substantial portions of the Software.

  THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED,
  INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A
  PARTICULAR PURPOSE AND NON-INFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT
  HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF
  CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE
  OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.

}}