'' =================================================================================================
''
''   File....... jm_nextion_demo.spin
''   Purpose....
''   Author..... Jon McPhalen
''               Copyright (c) 2022 Jon McPhalen
''               -- see below for terms of use
''   E-mail..... jon.mcphalen@gmail.com
''   Started....
''   Updated.... 23 DEC 2022
''
''   {$P1}
''
'' =================================================================================================


con { timing }

  _clkmode = xtal1 + pll16x
  _xinfreq = 5_000_000                                          ' use 5MHz crystal

  CLK_FREQ = (_clkmode >> 6) * _xinfreq                         ' system freq as a constant
  MS_001   = CLK_FREQ / 1_000                                   ' ticks in 1ms
  US_001   = CLK_FREQ / 1_000_000                               ' ticks in 1us


con { terminal }

  BR_TERM  = 115_200                                            ' terminal baud rate
  BR_HMI   =  57_600                                            ' nextion baud rate


con { fixed io pins }

  RX_PGM   = 31  { I }                                          ' serial / programming
  TX_PGM   = 30  { O }

  EE_SDA   = 29  { I/O }                                        ' i2c / eeprom
  EE_SCL   = 28  { I/O }

  LED_27   = 27  { O }                                          ' PAB / FLiP LEDs
  LED_26   = 26  { O }


con { app io }

  RX_HMI   = 15  { I }                                          ' Nextion HMI
  TX_HMI   = 14  { O }


con

  #true,  ON, OFF
  #false, NO, YES

  HMI_BFSIZE = 96

  SHOW_DETAILS = YES


obj

' main                                                          ' * master Spin cog
  time : "jm_time_80"                                           '   timing and delays (80MHz)
  prng : "jm_prng"                                              '   pseudo-random number generation
  io   : "jm_io"                                                '   simple io
  term : "jm_fullduplexserial"                                  ' * serial IO for terminal
  hmi  : "jm_nextion"                                           ' * Nextion coms

' * uses cog when loaded


var { globals }

  long  hashmi
  long  hmiver
  long  hmivalue
  long  last
  long  page

  byte  hmibuffer[HMI_BFSIZE]


pub main | len

  setup
  configure_hmi

  if (SHOW_DETAILS)
    show_nextion_details

  if (hashmi)
    repeat
      len := hmi.check_msg(@hmibuffer, HMI_BFSIZE-1)            ' message rx'd?
      if (len > 0)                                              ' if yes
        term.str(@hmibuffer)                                    '  show it
        process_hmi_message(@hmibuffer)                         '  deal with it

      if (newsec)                                               ' update nextion?
        update_nextion_time
        newsec := false

  repeat
    waitcnt(0)


pub process_hmi_message(p_msg)

  hmi.extract_cmd(p_msg)                                        ' get new message
  hmivalue := hmi.extract_value(p_msg)                          ' get value from message

  if strcomp(hmi.last_cmd, string("page"))                      ' new page?
    page := hmivalue
    return

  if strcomp(hmi.last_cmd, string("clock"))                     ' user set clock?
    set_rtc(hmivalue)
    return


pub set_rtc(nt)

'' Set RTC from Nextion command
'' -- time returned as decimal value in form hhmmssd

  rtcrun := false                                               ' stop bkg clock
  time.pause(1)

  day := nt //  10                                              ' extract rtc elements
  nt /=  10
  scs := nt // 100
  nt /= 100
  mns := nt // 100
  hrs := nt  / 100

  rtcrun := true                                                ' restart clock

  term.str(string("--> internal clock reset by Nextion", 13))


pub update_nextion_time | work, newtime

  if (page <> 0)
    return

  bytemove(@work, @scs, 4)                                      ' take snapshot of rtc

  newtime := work.byte[2]                                       ' hrs
  newtime := (newtime * 100) + work.byte[1]                     ' mns
  newtime := (newtime * 100) + work.byte[0]                     ' scs
  newtime := (newtime * 10)  + work.byte[3]                     ' day

  hmi.set_value(string("rawtime.val"), newtime)                 ' send time to Nextion
  hmi.click(string("hsSetClock"), 1)                            ' update display


pub show_nextion_details | s[6], n

'' Show details about Nextion display and internal user code

  wait_for_terminal(true)
  time.pause(250)

' term.fstr1(string("%s\r\r"), @hmibuffer)                      ' raw connection response

  if (hashmi == false)
    term.str(string("Failed to connect to Nextion."))
    repeat
      waitcnt(0)

  else
    term.fstr0(string("Connected\r"))

    hmi.get_model(@hmibuffer, @s)
    term.fstr1(string("-- Model........... %s\r"), @s)

    n := hmi.get_firmware(@hmibuffer)
    term.fstr1(string("-- Firmware........ %.2f\r"), n)

    if (hmi.get_touch(@hmibuffer))
      bytemove(@s, string("Yes"), 4)
    else
      bytemove(@s, string("No"), 3)
    term.fstr1(string("-- Touch Screen.... %s\r"), @s)

    n := hmi.get_flash_size(@hmibuffer)
    term.fstr1(string("-- Flash Size...... %d\r"), n)

    hmi.get_serial_number(@hmibuffer, @s)
    term.fstr1(string("-- Serial Number... %s\r"), @s)

    term.fstr3(string("-- User Code....... %d.%d.%d\r\r"), {
    } hmiver/100, hmiver//100/10, hmiver//10)


pub setup

'' Setup IO and objects for application

  time.start                                                    ' setup timing & delays

  prng.seed(-cnt, -cnt ~> 2, $CAFE_BABE, cnt <- 2, cnt)         ' seed randomizer

  io.start(0, 0)                                                ' clear all pins (master cog)

  term.tstart(BR_TERM)                                          ' start serial for terminal *

  rtcrun := true
  bgcog := cognew(rtc, @bgstack) + 1                            ' start rtc cog


pub configure_hmi | check

  hmi.start(RX_HMI, TX_HMI, BR_HMI)
  time.pause(25)

  hmi.connect(@hmibuffer)
  if (hmi.has_connection(@hmibuffer))
    hashmi := true
    hmi.reset(0, 0)
    page := 0
    hmiver := hmi.get_value(string("version"))
  else
    hashmi := false
    page := -1


pub wait_for_terminal(clear)

'' Wait for terminal to be open and key pressed

  term.rxflush
  term.rx
  if (clear)
    term.tx(term#CLS)


con { background rtc }

var

  long  bgcog
  long  bgstack[32]

  long  rtcrun
  long  newsec

  word  millis

  byte  scs
  byte  mns
  byte  hrs
  byte  day


pri rtc | t

  day := 1                                                      ' sunday

  t := cnt
  repeat
    waitcnt(t += MS_001)
    update_rtc
    heartbeat


pri update_rtc

  if (rtcrun == false)
    millis := 0
    return

  if (++millis == 1000)
    millis := 0
    if (++scs == 60)
      scs := 0
      if (++mns == 60)
        mns := 0
        if (++hrs == 24)
          hrs := 23
          if (++day == 8)
            day := 1

  if (millis == 0)
    newsec := true


pri heartbeat

  if (millis < 500)
    io.low(LED_26)
  else
    io.high(LED_26)


con { license }

{{

  Terms of Use: MIT License

  Permission is hereby granted, free of charge, to any person obtaining a copy of this
  software and associated documentation files (the "Software"), to deal in the Software
  without restriction, including without limitation the rights to use, copy, modify,
  merge, publish, distribute, sublicense, and/or sell copies of the Software, and to
  permit persons to whom the Software is furnished to do so, subject to the following
  conditions:

  The above copyright notice and this permission notice shall be included in all copies
  or substantial portions of the Software.

  THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED,
  INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A
  PARTICULAR PURPOSE AND NON-INFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT
  HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF
  CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE
  OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.

}}